"""Docs generation module."""

import importlib
from pathlib import Path

from jinja2 import Environment as JinjaEnvironment
from jinja2 import FileSystemLoader


def add_plot(name: str) -> str:
    """Add the plot to the code."""
    with (Path(__file__).parents[0] / "examples" / f"{name}.py").open() as file:
        code = file.read()
    chart = importlib.import_module(f"scripts.examples.{name}").chart
    if "png" in name:
        output_dir = Path(__file__).parents[1] / "docs" / "plots"
        output_dir.mkdir(exist_ok=True)
        chart.save(str(output_dir / f"{name}.png"))
        embedded_plot = f"![Plot](plots/{name}.png)\n"
    else:
        embedded_plot = f"```vegalite\n{chart.to_json(indent=2)}\n```\n"
    return f"```python\n{code}\n```\n\n{embedded_plot}"


if __name__ == "__main__":
    jinja_env = JinjaEnvironment(
        loader=FileSystemLoader(Path(__file__).parents[1] / "docs" / "templates")
    )

    for name_ in jinja_env.list_templates():
        template = jinja_env.get_template(name_)
        render = template.render(
            **{
                f.name.replace(".py", ""): add_plot(f.name.replace(".py", ""))
                for f in (Path(__file__).parents[0] / "examples").glob("*.py")
                if f.name.startswith(name_[0 : name_.find(".")])
            }
        )
        with Path(
            Path(__file__).parents[1] / "docs" / name_.replace(".jinja", "")
        ).open(mode="w") as f:
            f.write(render)



"""**Custom visualization themes**.

Here we define a custom theme for
`Altair <https://altair-viz.github.io/index.html>`_ and
`Matplotlib <https://matplotlib.org/stable/index.html>`_.
We export a :data:`~.theme` object that should be used across the code
to activate and deactivate the custom theme:

.. doctest::

    >>> from driviz import theme

    >>> theme.enable()

This theme uses Google's
`Roboto <https://fonts.google.com/specimen/Roboto>`_ font,
**it should be installed in your OS**.
You can install it in Ubuntu with:

.. code-block:: console

    $ sudo apt install fonts-roboto

The following example generates the interactive plot
``data/results/altair_example_barh.html``:

.. doctest::

    >>> import altair as alt
    >>> import numpy as np
    >>> import pandas as pd
    >>> import random

    >>> from driviz import theme

    >>> theme.enable()

    >>> variety =  [f"V{i}" for i in range(10)]
    >>> site = [f"site{i:02d}" for i in range(14)]
    >>> k = 10000
    >>> df = pd.DataFrame(
    ...     data={
    ...         "yield": np.random.rand(k,),
    ...         "variety": random.choices(variety, k=k),
    ...         "site": random.choices(site, k=k),
    ...     }
    ... )

    >>> selection = alt.selection_point(fields=["site"], bind="legend")

    >>> bars = (
    ...     alt.Chart(df)
    ...     .mark_bar()
    ...     .encode(
    ...         x=alt.X("sum(yield):Q", stack="zero"),
    ...         y=alt.Y("variety:N"),
    ...         color=alt.Color("site"),
    ...         opacity=alt.condition(
    ...             selection, alt.value(1), alt.value(0.2)
    ...         )
    ...     )
    ...     .properties(title="Example chart")
    ...     .add_params(selection)
    ... )

    >>> text = (
    ...     alt.Chart(df)
    ...     .mark_text(dx=-15, dy=3, color="white")
    ...     .encode(
    ...         x=alt.X("sum(yield):Q", stack="zero"),
    ...         y=alt.Y("variety:N"),
    ...         detail="site:N",
    ...         text=alt.Text("sum(yield):Q", format=".1f")
    ...     )
    ... )

    >>> chart = bars + text
    >>> chart.save(
    ...     "altair_example_barh.html"
    ... )

For more examples, visit the
`Example Gallery <https://altair-viz.github.io/gallery/index.html>`_.

"""

from copy import deepcopy
from typing import Any, Literal, no_type_check

import altair as alt
import matplotlib as mpl
from cycler import cycler
from pydantic import BaseModel, ConfigDict
from pydantic_extra_types.color import Color

WHICH_PACKAGE = Literal["all", "alt", "mpl"]


class _Base(BaseModel):
    """Base class."""

    model_config = ConfigDict(
        validate_default=True,
    )


class CategoryPalettes(_Base):
    """Discrete color palettes for categories."""

    palette: list[Color] = [
        Color("#00004E"),
        Color("#0043AF"),
        Color("#3D97F2"),
        Color("#73BAFF"),
        Color("#A8C2AA"),
        Color("#CBC771"),
        Color("#FFCD1B"),
        Color("#FF9424"),
        Color("#FF6D2A"),
        Color("#FF3333"),
        Color("#A0213E"),
        Color("#601445"),
    ]


class CustomColors(_Base):
    """Custom theme colors."""

    dark_blue: Color = Color("#00004E")
    blue: Color = Color("#0043AF")
    light_blue: Color = Color("#73BAFF")
    grey: Color = Color("#DFE6EA")
    red: Color = Color("#FF3333")
    yellow: Color = Color("#FFCD1B")
    black: Color = Color("#000000")
    white: Color = Color("#fff")


class VegaActions(_Base):
    """Vega actions options."""

    export: bool = True
    """Export as PNG and SVG."""
    source: bool = False
    """View the Vega source code."""
    compiled: bool = False
    """View the Vega compiled code."""
    editor: bool = False
    """Open in Vega Editor."""


class Theme(_Base):
    """Custom Altair visualization theme.

    This class defines a custom Altair theme and provides methods
    to :meth:`~.enable` and :meth:`~.disable` it.

    """

    _default_theme: str = "default"
    _default_renderer: str = "default"
    _default_mpl_params: mpl.RcParams = deepcopy(mpl.rcParams)
    """Altair's default theme name."""
    theme_name: str = "dribia"
    """Name under which the theme is registered."""
    color: CustomColors = CustomColors()
    """Custom colors."""
    category_palettes: CategoryPalettes = CategoryPalettes()
    """Category palettes."""
    font: str = "Roboto"
    """Custom font."""
    font_weight: str = "normal"
    """Custom font weight."""
    legend_font_weight: str = "normal"
    """Custom font weight legend."""
    font_size_base: int = 18
    """Custom font size."""
    font_size_xl: float = font_size_base * 1.5
    """Custom font large size."""
    font_size_sm: float = font_size_base * 0.9
    """Custom font small size."""
    font_color: Color = color.dark_blue
    """Custom font color."""
    height: int = 400
    """Custom window height."""
    width: int = 711
    """Custom window width."""
    dpi: int = 50
    """Custom DPI."""
    actions: VegaActions = VegaActions()
    """Actions displayed in the actions menu."""
    locale: str = "en-GB"
    """
    Locale options for the theme (ISO 639-1 two-letter codes). It works for Altair only.
    Available options in: https://github.com/d3/d3-format/tree/main/locale
    """

    def _get_alt_theme(self) -> dict[str, Any]:
        """Build and get the theme's configuration dictionary.

        Returns: Theme's configurations.

        """
        _hex_categories_palette = [c.as_hex() for c in self.category_palettes.palette]
        return {
            "config": {
                "background": "white",
                "view": {
                    "continuousHeight": self.height,
                    "continuousWidth": self.width,
                },
                "title": {
                    "fontSize": self.font_size_xl,
                    "anchor": "start",
                    "color": self.font_color.as_hex(),
                    "font": self.font,
                    "fontWeight": self.font_weight,
                    "offset": 15,
                    "subtitleFont": self.font,
                    "subtitleFontWeight": self.font_weight,
                    "subtitleFontSize": self.font_size_sm,
                    "subtitleFontStyle": self.legend_font_weight,
                },
                "axis": {
                    "titleFont": self.font,
                    "titleColor": self.font_color.as_hex(),
                    "titleFontSize": self.font_size_sm,
                    "titleFontWeight": self.font_weight,
                    "labelFont": self.font,
                    "labelColor": self.font_color.as_hex(),
                    "labelFontSize": self.font_size_base,
                    "labelFontWeight": self.legend_font_weight,
                    "grid": True,
                    "gridColor": self.color.grey.as_hex(),
                    "gridOpacity": 1,
                    "domain": False,
                    "tickColor": self.font_color.as_hex(),
                },
                "legend": {
                    "titleFont": self.font,
                    "titleColor": self.font_color.as_hex(),
                    "titleFontSize": self.font_size_base,
                    "titleFontWeight": self.font_weight,
                    "labelFont": self.font,
                    "labelColor": self.font_color.as_hex(),
                    "labelFontSize": self.font_size_sm,
                    "labelFontWeight": self.legend_font_weight,
                },
                "header": {
                    "titleFont": self.font,
                    "titleColor": self.font_color.as_hex(),
                    "titleFontSize": self.font_size_base,
                    "titleFontWeight": self.font_weight,
                    "labelFont": self.font,
                    "labelColor": self.font_color.as_hex(),
                    "labelFontSize": self.font_size_base,
                    "labelFontWeight": self.font_weight,
                },
                "line": {"stroke": self.color.blue.as_hex()},
                "circle": {"fill": self.color.blue.as_hex(), "stroke": None},
                "point": {
                    "stroke": self.color.blue.as_hex(),
                    "fill": self.color.light_blue.as_hex(),
                },
                "rect": {"fill": self.color.blue.as_hex()},
                "range": {
                    "category": _hex_categories_palette,
                    "ramp": _hex_categories_palette,
                    "diverging": [
                        self.color.blue.as_hex(),
                        self.color.white.as_hex(),
                        self.color.red.as_hex(),
                    ],
                    "heatmap": _hex_categories_palette[0:10],
                },
            }
        }

    @staticmethod
    def _get_cmap(_hex_categories_palette: list[str]) -> str:
        """Build and get the theme's color map."""
        if "dribia" not in mpl.colormaps:
            mpl.colormaps.register(
                cmap=mpl.colors.LinearSegmentedColormap.from_list(
                    name="dribia", colors=_hex_categories_palette[0:10]
                ),
                name="dribia",
            )
        return "dribia"

    def _get_mpl_theme(self) -> dict[str, Any]:
        """Build and get the Matplotlib theme's configuration dict.

        Returns: Matplotlib theme's configurations.

        """
        _hex_categories_palette = [c.as_hex() for c in self.category_palettes.palette]
        return {
            "axes.prop_cycle": cycler("color", _hex_categories_palette),
            "axes.edgecolor": self.color.grey.as_hex(),
            "axes.grid": True,
            "axes.labelcolor": self.font_color.as_hex(),
            "axes.labelsize": self.font_size_xl,
            "axes.labelweight": "bold",
            "axes.titlecolor": self.font_color.as_hex(),
            "axes.titlelocation": "left",
            "axes.titlepad": 15.0,
            "axes.titlesize": self.font_size_xl,
            "axes.titleweight": "normal",
            "figure.dpi": self.dpi,
            "figure.figsize": [self.width / self.dpi, self.height / self.dpi],
            "figure.titlesize": self.font_size_xl,
            "font.family": self.font,
            "font.size": self.font_size_base,
            "font.weight": "normal",
            "grid.color": self.color.grey.as_hex(),
            "image.cmap": self._get_cmap(_hex_categories_palette),
            "legend.fontsize": self.font_size_base,
            "legend.labelcolor": self.font_color.as_hex(),
            "legend.loc": "upper right",
            "legend.title_fontsize": self.font_size_base,
            "lines.color": self.color.blue.as_hex(),
            "patch.facecolor": self.color.blue.as_hex(),
            "text.color": self.font_color.as_hex(),
            "xtick.color": self.font_color.as_hex(),
            "ytick.color": self.font_color.as_hex(),
            "xtick.labelsize": self.font_size_base,
            "ytick.labelsize": self.font_size_base,
        }

    @no_type_check
    def enable(self, which: WHICH_PACKAGE = "all") -> None:
        """Enable the custom theme.

        1. If the theme is not yet registered in Altair, we register it.
        2. We enable the theme.
        3. We set the actions we want to show in the actions menu.
        4. We set the configurations for Matplotlib.

        Args:
            which: Graphics environment to enable the theme on.

        """
        if which in ["all", "alt"]:

            @alt.theme.register(self.theme_name, enable=True)
            def dribia_theme():
                return alt.theme.ThemeConfig(
                    **self._get_alt_theme()
                )  # pragma: no cover

            alt.renderers.set_embed_options(
                actions=self.actions.model_dump(),
                time_format_locale=self.locale,
                format_locale=self.locale,
            )

        if which in ["all", "mpl"]:
            mpl.rcParams.update(self._get_mpl_theme())

    @no_type_check
    def disable(self, which: WHICH_PACKAGE = "all") -> None:
        """Disable the custom theme.

        Disabling the theme is just enabling back the default one.

        Args:
            which: Graphics environment to disable the theme on.

        """
        if which in ["all", "alt"]:
            alt.theme.enable(self._default_theme)
            alt.renderers.enable(self._default_renderer)
        if which in ["all", "mpl"]:
            mpl.rcParams.update(self._default_mpl_params)

    @no_type_check
    def set_basic_colors(self, six: bool = False, first_grey: bool = False) -> None:
        """Set the theme's configuration with four basic colors.

        Args:
            six: whether to set the palette to six colors, otherwise, it
              will be set up with four colors.
            first_grey: whether to set the first color to grey.

        .. doctest::

            >>> import altair as alt
            >>> from driviz import theme

            >>> theme.set_basic_colors()

        """
        if six:
            colors = [
                Color("#00004E"),
                Color("#3D97F2"),
                Color("#A8C2AA"),
                Color("#FFCD1B"),
                Color("#FF6D2A"),
                Color("#A0213E"),
            ]
            name = "basic_colors_six"
        else:
            colors = [
                Color("#00004E"),
                Color("#73BAFF"),
                Color("#FFCD1B"),
                Color("#FF3333"),
            ]
            name = "basic_colors"
        if first_grey:
            colors = [Color("#BABFC2"), *colors]
            name += "_grey"

        new_theme = self._get_alt_theme()
        new_theme["config"]["range"]["category"] = [c.as_hex() for c in colors]

        @alt.theme.register(name, enable=True)
        def dribia_basic_colors_theme():
            return alt.theme.ThemeConfig(**new_theme)  # pragma: no cover

        alt.renderers.set_embed_options(
            actions=self.actions.model_dump(),
            time_format_locale=self.locale,
            format_locale=self.locale,
        )

        mpl_theme = self._get_mpl_theme()
        mpl_theme["axes.prop_cycle"] = cycler("color", [c.as_hex() for c in colors])
        mpl.rcParams.update(mpl_theme)


theme = Theme()
"""This is the object that should be used to enable and disable the
theme across the code.

.. doctest::

    >>> from driviz import theme

    >>> theme.enable()

:meta hide-value:
"""

__all__ = ["Theme", "theme"]



"""Test the module initialization and version information."""

import sys
from importlib import reload
from importlib.metadata import PackageNotFoundError
from unittest.mock import patch


class TestInit:
    """Test the module initialization and version information."""

    @patch("importlib.metadata.version", side_effect=PackageNotFoundError)
    def test_version_not_found(self, mock_version):
        """Test `__version__` when the package is not found."""
        if "driviz" in sys.modules:
            del sys.modules["driviz"]

        import driviz

        reload(driviz)
        assert driviz.__version__ == "unknown"

    @patch("importlib.metadata.version")
    def test_version_found(self, mock_version):
        """Test `__version__` when the package is found."""
        mock_version.return_value = "0.1.0"
        if "driviz" in sys.modules:
            del sys.modules["driviz"]

        import driviz

        reload(driviz)
        assert driviz.__version__ == "0.1.0"



import altair as alt
from pydantic_extra_types.color import Color
from vega_datasets import data  # type: ignore[import-untyped]

from driviz.theme import Theme

theme = Theme(font="Arial", font_color=Color("green"))
theme.set_basic_colors()

source = data.iowa_electricity()
chart = (
    alt.Chart(source)
    .mark_area()
    .encode(
        x="year:T",
        y=alt.Y("net_generation:Q").stack("normalize"),
        color="source:N",
    )
)



import random

import altair as alt
import numpy as np
import pandas as pd

from driviz import theme

alt.data_transformers.disable_max_rows()

theme.enable()

variety = [f"V{i}" for i in range(10)]
site = [f"site{i:02d}" for i in range(14)]
k = 10000
rng = np.random.default_rng()
random_df = pd.DataFrame(
    data={
        "yield": rng.random(k),
        "variety": random.choices(variety, k=k),
        "site": random.choices(site, k=k),
    }
)
selection = alt.selection_point(fields=["site"], bind="legend")
bars = (
    alt.Chart(random_df)
    .mark_bar()
    .encode(
        x=alt.X("sum(yield):Q", stack="zero"),
        y=alt.Y("variety:N", title="Variety"),
        color=alt.Color("site", title="Site"),
        opacity=alt.condition(selection, alt.value(1), alt.value(0.2)),
    )
    .add_params(selection)
)
text = (
    alt.Chart(random_df)
    .mark_text(dx=-15, dy=3, color="white")
    .encode(
        x=alt.X("sum(yield):Q", stack="zero"),
        y=alt.Y("variety:N"),
        detail="site:N",
        text=alt.Text("sum(yield):Q", format=".1f"),
    )
)
chart = (bars + text).properties(title="Example chart", height=300)



import altair as alt
from vega_datasets import data  # type: ignore[import-untyped]

from driviz.theme import Theme

theme = Theme(locale="ca-ES")  # Set Catalan language
theme.set_basic_colors()

source = data.stocks().sort_values(by="date").iloc[-100:]
chart = (
    alt.Chart(source)
    .mark_line()
    .encode(
        x=alt.X("date:T", title="Data"),
        y=alt.Y("price:Q", title="Preu"),
        color=alt.Color("symbol:N", title="Símbol"),
    )
    .properties(title="Estocs", width=600, height=300)
)



import altair as alt
import pandas as pd

from driviz import theme

theme.enable()

source = pd.DataFrame({"values": [12, 23, 47, 6, 52, 19]})
base = alt.Chart(source).encode(
    alt.Theta("values:Q").stack(True),  # noqa: PD013
    alt.Radius("values").scale(type="sqrt", zero=True, rangeMin=20),
    color="values:N",
)
c1 = base.mark_arc(innerRadius=20, stroke="#fff")
c2 = base.mark_text(radiusOffset=10).encode(text="values:Q")
chart = c1 + c2



import altair as alt
import pandas as pd

from driviz import theme

theme.enable()

source = pd.DataFrame(
    data={"category": [1, 2, 3, 4, 5, 6], "value": [4, 6, 10, 3, 7, 8]},
)
color_mapping = {
    1: "#FD0100",
    2: "#F76915",
    3: "#EEDE04",
    4: "#A0D636",
    5: "#2FA236",
    6: "#333ED4",
}
chart = (
    alt.Chart(source)
    .mark_arc()
    .encode(
        theta="value",
        color=alt.Color(
            "category:N",
            title="Category",
            scale=alt.Scale(
                domain=list(color_mapping),
                range=list(color_mapping.values()),
            ),
        ),
    )
)



import altair as alt
import pandas as pd

from driviz import theme

theme.set_basic_colors(six=True)

source = pd.DataFrame(
    data={"category": [1, 2, 3, 4, 5, 6], "value": [4, 6, 10, 3, 7, 8]},
)
chart = (
    alt.Chart(source)
    .mark_arc()
    .encode(theta="value", color=alt.Color("category:N", title="Category"))
)



"""Viz theme test module."""

from typing import Any

import altair as alt
import matplotlib as mpl

from driviz import theme

_default_mpl_params = mpl.rcParams.copy()
_default_alt_theme = alt.theme.active


def _compare_mpl_params(p1: dict[str, Any], p2: dict[str, Any]) -> bool:
    """Utility to compare matplotlib params excluding the backend."""
    _p1 = dict(p1)
    _p2 = dict(p2)
    _p1.pop("backend")
    _p2.pop("backend")
    return _p1 == _p2


def test_theme():
    """Test the enable method from Theme object."""
    assert isinstance(theme._get_mpl_theme(), dict)
    assert isinstance(theme._get_alt_theme(), dict)
    assert _compare_mpl_params(mpl.rcParams.copy(), _default_mpl_params)
    assert alt.theme.active == _default_alt_theme
    theme.enable("alt")
    assert _compare_mpl_params(mpl.rcParams.copy(), _default_mpl_params)
    assert alt.theme.active == theme.theme_name
    theme.enable("mpl")
    assert alt.theme.active == theme.theme_name
    assert not _compare_mpl_params(mpl.rcParams.copy(), _default_mpl_params)
    theme.disable("alt")
    assert alt.theme.active == _default_alt_theme
    assert not _compare_mpl_params(mpl.rcParams.copy(), _default_mpl_params)
    theme.disable("mpl")
    assert _compare_mpl_params(mpl.rcParams.copy(), _default_mpl_params)
    assert alt.theme.active == _default_alt_theme
    theme.enable("all")
    assert alt.theme.active == theme.theme_name
    assert not _compare_mpl_params(mpl.rcParams.copy(), _default_mpl_params)
    theme.disable("all")
    assert _compare_mpl_params(mpl.rcParams.copy(), _default_mpl_params)
    assert alt.theme.active == _default_alt_theme
    theme.set_basic_colors()
    assert alt.theme.active == "basic_colors"
    theme.set_basic_colors(six=True)
    assert alt.theme.active == "basic_colors_six"
    theme.set_basic_colors(first_grey=True)
    assert alt.theme.active == "basic_colors_grey"
    theme.set_basic_colors(six=True, first_grey=True)
    assert alt.theme.active == "basic_colors_six_grey"



import altair as alt
from vega_datasets import data  # type: ignore[import-untyped]

from driviz import theme

theme.set_basic_colors()

source = data.iowa_electricity()
chart = (
    alt.Chart(source)
    .mark_area()
    .encode(
        x="year:T",
        y=alt.Y("net_generation:Q").stack("normalize"),
        color="source:N",
    )
)
